@@ -24,6 +24,11 @@ from torch import nn
 from torch.nn import functional as F
 from torch.utils import data
 from torchvision import transforms
+from torchstat import stat
+
+
+def compute_num_params(net, shape):
+    stat(net, shape)
 
 
 #mylib = sys.modules[__name__]
@@ -394,7 +399,7 @@ def train_classify(net, train_iter, val_iter, loss, num_epochs, updater, device=
     # animator = Animator(xlabel='epoch', xlim=[1, num_epochs], ylim=[0, 1],
     #                     legend=['train loss', 'train acc', 'test acc'])
 
-    animator = Animator(xlabel='epoch', xlim=[1, num_epochs], ylim=None,
+    animator = Animator(xlabel='epoch', xlim=[1, num_epochs], ylim=[0, 1],
                     legend=['train loss', 'train acc', 'test acc'])
 
     train_epoch_loss, train_epoch_acc, val_epoch_acc = [], [], []
@@ -506,7 +511,7 @@ def train_regression(net, train_iter, val_iter, loss, num_epochs, updater, devic
             for model in net.children():
                 handles.append(model.register_full_backward_hook(monitor_grad_hook))
 
-    animator = Animator(xlabel='epoch', xlim=[1, num_epochs], ylim=None,
+    animator = Animator(xlabel='epoch', xlim=[1, num_epochs], ylim=[0, 1],
                         legend=['train loss', 'val loss'])
     
     net.to(device)
@@ -976,20 +981,6 @@ class DotProductAttention(nn.Module):
         return torch.bmm(self.dropout(self.attention_weights), values)#batch*num_query*值的维度
 
 
-# class PositionalEncoding(nn.Module):
-#     def __init__(self, emb_size, dropout, max_seq_len=1000):
-#         super().__init__()
-#         self.dropout = nn.Dropout(dropout)
-#         self.P = torch.zeros((1, max_seq_len, emb_size))
-#         X = torch.arange(max_seq_len, dtype=torch.float32).reshape(
-#             -1, 1) / torch.pow(10000, torch.arange(
-#             0, emb_size, 2, dtype=torch.float32) / emb_size)
-#         self.P[:, :, 0::2] = torch.sin(X)
-#         self.P[:, :, 1::2] = torch.cos(X)
-#     def forward(self, x):
-#         x = x + self.P[:, :x.shape[1], :].to(x.device)
-#         return self.dropout(x)
-
 class PositionalEncoding(nn.Module):
     """位置编码"""
     def __init__(self, num_hiddens, dropout, max_len=1000):
@@ -1050,7 +1041,7 @@ def transpose_output(X, num_heads):
     return X.reshape(X.shape[0], X.shape[1], -1)
 
 
-#--------------------------------------------------------------------------------
+#--------------------------------------------动态学习率------------------------------------
 # 预热+余弦学习率调度器
 class CosineScheduler:
     '''会在epoch==warmup_steps时达到base_lr'''
@@ -1078,7 +1069,7 @@ class CosineScheduler:
         return self.base_lr
 
 
-#---------------------------------------------------------------------------------
+#---------------------------------------目标检测相关------------------------------------------
 def box_corner_to_center(boxes):
     """从（左上，右下）转换到（中间，宽度，高度）"""
     x1, y1, x2, y2 = boxes[:, 0], boxes[:, 1], boxes[:, 2], boxes[:, 3]
@@ -1104,4 +1095,213 @@ def bbox_to_rect(bbox, color):
     # ((左上x,左上y),宽,高)
     return plt.Rectangle(
         xy=(bbox[0], bbox[1]), width=bbox[2]-bbox[0], height=bbox[3]-bbox[1],
-        fill=False, edgecolor=color, linewidth=2)
\ No newline at end of file
+        fill=False, edgecolor=color, linewidth=2)
+
+
+def multibox_prior(img, sizes, ratios):
+    """生成以每个像素为中心具有不同形状的锚框"""
+    in_height, in_width = img.shape[-2:]
+    device, num_sizes, num_ratios = img.device, len(sizes), len(ratios)
+    boxes_per_pixel = (num_sizes + num_ratios - 1)
+    size_tensor = torch.tensor(sizes, device=device)
+    ratio_tensor = torch.tensor(ratios, device=device)
+    # 生成锚框的所有中心点
+    # 这里以0 1 2为例，中心点是0.5和1.5所以要加上0.5, /in_height是为了进行归一化
+    center_h = (torch.arange(in_height, device=device) + 0.5) / in_height
+    center_w = (torch.arange(in_width, device=device) + 0.5) / in_width
+    shift_y, shift_x = torch.meshgrid(center_h, center_w, indexing='ij')
+    shift_y, shift_x = shift_y.reshape(-1), shift_x.reshape(-1)#此时shift_y和shift_x一对一地形成了所有中心点下标
+    # 生成“boxes_per_pixel”个高和宽，
+    # 之后用于创建锚框的四角坐标(xmin,xmax,ymin,ymax)
+    # w为一行，其中每个元素为不同的锚框的宽度
+    w = torch.cat((sizes[0] * torch.sqrt(in_height * ratio_tensor[:] / in_width),
+                     size_tensor[1:] * torch.sqrt(in_height * ratio_tensor[0] / in_width)))
+    # h为一行，其中每个元素为不同的锚框的高度
+    h = torch.cat((sizes[0] * torch.sqrt(in_width / ratio_tensor[:] / in_height), 
+                     size_tensor[1:] * torch.sqrt(in_width / ratio_tensor[0] / in_height)))
+    # 除以2来获得半高和半宽
+    anchor_manipulations = torch.stack((-w, -h, w, h)).T.repeat(in_height * in_width, 1) / 2
+    # 每个中心点都将有“boxes_per_pixel”个锚框，
+    # 所以生成含所有锚框中心的网格，重复了“boxes_per_pixel”次
+    out_grid = torch.stack([shift_x, shift_y, shift_x, shift_y],
+                dim=1).repeat_interleave(boxes_per_pixel, dim=0)
+    output = out_grid + anchor_manipulations
+    # 只能返回一张图片的所有锚框
+    return output.unsqueeze(0) # 1 * 锚框个数 * 4(左上和右下下标)
+
+
+
+# 显示以图像中一个像素为中心的所有锚框
+def show_bboxes(axes, bboxes, labels=None, colors=None):
+    """显示所有边界框"""
+    def _make_list(obj, default_values=None):
+        if obj is None:
+            obj = default_values
+        elif not isinstance(obj, (list, tuple)):
+            obj = [obj]
+        return obj
+    
+    labels = _make_list(labels)
+    colors = _make_list(colors, ['b','g','r','m','c'])
+    for i, bbox in enumerate(bboxes):
+        color = colors[i % len(colors)]
+        rect = bbox_to_rect(bbox.detach().numpy(),color)
+        axes.add_patch(rect)
+        if labels and len(labels) > i:
+            text_color = 'k' if color == 'w' else 'w'
+            axes.text(rect.xy[0], rect.xy[1], labels[i], va='center',
+                     ha='center', fontsize=9, color=text_color,
+                     bbox=dict(facecolor=color, lw=0))
+
+
+# 交并比(IoU)
+def box_iou(boxes1,boxes2):
+    # 左上右下形式坐标
+    # boxes1：(boxes1的数量,4),
+    # boxes2：(boxes2的数量,4),
+    """计算两个锚框或边界框列表中成对的交并比"""
+    box_area = lambda boxes: ((boxes[:,2] - boxes[:,0]) *
+                             (boxes[:,3] - boxes[:,1]))
+    areas1 = box_area(boxes1) # 锚框1的面积
+    areas2 = box_area(boxes2) # 锚框2的面积
+    inter_upperlefts = torch.max(boxes1[:,None,:2],boxes2[:,:2]) 
+    inter_lowerrights = torch.min(boxes1[:,None,2:],boxes2[:,2:])
+    inters = (inter_lowerrights - inter_upperlefts).clamp(min=0)
+    # inter_areasand and union_areas的形状:(boxes1的数量,boxes2的数量)
+    inter_areas = inters[:,:,0] * inters[:,:,1] # 交集的面积
+    union_areas = areas1[:,None] + areas2 - inter_areas # 并集的面积
+    return inter_areas / union_areas # num_box1 * num_box2
+
+
+
+# 将真实边界框分配给锚框
+def assign_anchor_to_bbox(anchors,ground_truth,device,iou_threshold=0.5):
+    """将最接近的真实边界框分配给锚框"""
+    num_anchors, num_gt_boxes = anchors.shape[0], ground_truth.shape[0]
+    jaccard = box_iou(anchors,ground_truth) # 计算所有的锚框和真实边缘框的IOU
+    anchors_bbox_map = torch.full((num_anchors,), -1, dtype=torch.long, device=device)    
+    max_ious, indices = torch.max(jaccard, dim=1)#indices为每一个锚框对应的真实的边界框的标号, max_ious为每一个锚框对应的真实边界框的iou
+    anc_i = torch.nonzero(max_ious >= iou_threshold).reshape(-1) #找到所有iou值>0.5的锚框的下标
+    box_j = indices[max_ious >= iou_threshold]#找到所有iou值>0.5的锚框对应的真实边界框
+    anchors_bbox_map[anc_i] = box_j
+    #到现在上面的每一个iou>=0.5的锚框都分配了一个真实的边界框，但是此时并不一定每个真实边界框都分配到了一个锚框，
+    #所以下面的代码要对每一个真实边界框进行遍历,并未其分配一个锚框
+    col_discard = torch.full((num_anchors,),-1)
+    row_discard = torch.full((num_gt_boxes,),-1)
+    for _ in range(num_gt_boxes):
+        max_idx = torch.argmax(jaccard, dim=None) # 找IOU最大的锚框,找的是全局最大的锚框
+        box_idx = (max_idx % num_gt_boxes).long()
+        anc_idx = (max_idx / num_gt_boxes).long()
+        anchors_bbox_map[anc_idx] = box_idx
+        jaccard[:,box_idx] = col_discard # 把最大Iou对应的锚框在 锚框-类别 矩阵中的一列删掉
+        jaccard[anc_idx,:] = row_discard # 把最大Iou对应的锚框在 锚框-类别 矩阵中的一行删掉
+    return anchors_bbox_map#返回每一个分配了边界框的锚框对应的真实边界框的下标
+
+def offset_boxes(anchors, assigned_bb, eps=1e-6):
+    """对锚框偏移量的转换"""
+    c_anc = box_corner_to_center(anchors)
+    c_assigned_bb = box_corner_to_center(assigned_bb)
+    offset_xy = 10 * (c_assigned_bb[:, :2] - c_anc[:, :2]) / c_anc[:, 2:]
+    offset_wh = 5 * torch.log(eps + c_assigned_bb[:, 2:] / c_anc[:, 2:])
+    offset = torch.cat([offset_xy, offset_wh], axis=1)
+    return offset # 尽量使得 offset 让 machine learning 算法好预测
+
+
+# 标记锚框的类和偏移量
+def multibox_target(anchors, labels):
+    """使用真实边界框标记锚框"""
+    batch_size, anchors = labels.shape[0], anchors.squeeze(0)
+    batch_offset, batch_mask, batch_class_labels = [], [], []
+    device, num_anchors = anchors.device, anchors.shape[0]
+    for i in range(batch_size):
+        label = labels[i,:,:]
+        anchors_bbox_map = assign_anchor_to_bbox(anchors,label[:,1:],device)   
+        class_labels = torch.zeros(num_anchors, dtype=torch.long,device=device)  
+        assigned_bb = torch.zeros((num_anchors,4), dtype=torch.float32,device=device)   
+        indices_true =torch.nonzero(anchors_bbox_map >= 0)# 分配了真实边界框的锚框下标
+        bb_idx = anchors_bbox_map[indices_true] #分配了边界框的锚框对应的真实边界框的下标
+        class_labels[indices_true] = label[bb_idx,0].long() + 1 #所有锚框对应的类别，0表示背景，>0表示物体, (num_anchors,)
+        assigned_bb[indices_true] = label[bb_idx, 1:] #所有锚框对应的真实的边界框的坐标，num_anchors * 4
+        bbox_mask = ((anchors_bbox_map >= 0).float().unsqueeze(-1)).repeat(1,4) # num_anchors * 4,没有分配真实边界框的锚框对应4个0,分配的对应4个1   
+        offset = offset_boxes(anchors, assigned_bb) * bbox_mask#乘上bbox_mask是因为有的锚框对应的真实边界框坐标全是0，为了把这部分去掉, num_anchors*4
+        batch_offset.append(offset.reshape(-1))
+        batch_mask.append(bbox_mask.reshape(-1))
+        batch_class_labels.append(class_labels)
+    bbox_offset = torch.stack(batch_offset) #batch_size * (num_anchors*4)
+    bbox_mask = torch.stack(batch_mask) #batch_size * (num_anchors*4)
+    class_labels = torch.stack(batch_class_labels) #batch_size * num_anchors
+    # bbox_offset返回每一个锚框到真实标注框的offset偏移
+    # bbox_mask为0表示背景锚框，就不用了，>0表示对应真实的物体
+    # class_labels为锚框对应类的编号
+    return (bbox_offset, bbox_mask, class_labels)
+
+
+# 应用逆偏移变换来返回预测的边界框坐标
+def offset_inverse(anchors,offset_preds):
+    """根据带有预测偏移量的锚框来预测边界框"""
+    anc = box_corner_to_center(anchors)
+    pred_bbox_xy = (offset_preds[:,:2] * anc[:,2:] / 10) + anc[:,:2]
+    pred_bbox_wh = torch.exp(offset_preds[:,2:] / 5) * anc[:, 2:]
+    pred_bbox = torch.cat((pred_bbox_xy, pred_bbox_wh), axis=1)
+    predicted_bbox = box_center_to_corner(pred_bbox)
+    return predicted_bbox # 将锚框用偏移量进行偏移，得到预测的边界框
+
+
+def nms(boxes, scores, iou_threshold):
+    #boxes为所有的预测框， scores为每个预测框对应的预测类别概率最大的值
+    """对预测边界框的置信度进行排序"""
+    B = torch.argsort(scores, dim = -1, descending=True)#得到从大到小的预测概率所对应的预测框下标
+    keep = []
+    while B.numel()>0: # 直到把所有框都访问过了，再退出循环
+        i = B[0] # B中的预测概率最大值的下标（对应着相应的预测框）
+        keep.append(i)
+        if B.numel() == 1: break
+        # 所有的iou大于阈值的全部去掉    
+        iou = box_iou(boxes[i,:].reshape(-1,4),
+                     boxes[B[1:],:].reshape(-1,4)).reshape(-1)
+        inds = torch.nonzero(iou <= iou_threshold).reshape(-1) #得到所有<阈值的下标
+        B = B[inds + 1]#+1是因为得到这些与最大概率预测框的iou小于阈值的预测框的在B中的下标，然后把B更新成小于阈值的预测框
+    return torch.tensor(keep, dtype=torch.long, device=boxes.device)#返回未被抑制的预测框的下标
+
+
+# 将非极大值抑制应用于预测边界框
+def multibox_detection(cls_probs,offset_preds,anchors,nms_threshold=0.5,pos_threshold=0.009999999):
+    #cls_probs:每一个预测框对每一类的预测概率，batch * num_classes * num_predicted_bb
+    #offset_preds:batch * num_anchors * 4
+    device, batch_size = cls_probs.device, cls_probs.shape[0]
+    anchors = anchors.squeeze(0)
+    num_classes, num_anchors = cls_probs.shape[1], cls_probs.shape[2]
+    out = []
+    for i in range(batch_size):
+        cls_prob, offset_pred = cls_probs[i], offset_preds[i].reshape(-1,4) # num_anchors * 4 
+        predicted_bb = offset_inverse(anchors,offset_pred) # 把预测框拿出来
+        #下面conf得到不考率背景的每一个预测框的对每一类物体的最大概率值和相应的下标,classid这里变成用0代表第一类物体了
+        #不考虑背景是因为在抑制的时候防止因为背景框把物体检测的框抑制了
+        conf, class_id = torch.max(cls_prob ,dim = 0)
+        background_indices = torch.nonzero(class_id == 0).reshape(-1)# 背景预测框的下标
+        non_background_indices = torch.nonzero(class_id != 0).reshape(-1)# 非背景预测框的下标
+        non_back_predictbb = predicted_bb[non_background_indices]
+        non_back_conf = conf[non_background_indices]
+        keep = nms(non_back_predictbb, non_back_conf, nms_threshold)
+        # 获取被抑制的预测框
+        non_keep = []
+        for idx, _ in enumerate(non_back_predictbb):
+            if idx not in keep:
+                non_keep.append(idx)
+        # 将被抑制的预测框的类别设置为0
+        class_id[non_background_indices[non_keep]] = 0
+        # 下面就是把所有预测框重新排列了一下，前面放的没有被抑制的，后面放的被抑制的
+        all_sorted_indices = torch.cat((non_background_indices[keep],  #非背景框
+                                        non_background_indices[non_keep],#背景框
+                                        background_indices), dim=0)#背景框
+        class_id = class_id[all_sorted_indices]
+        conf = conf[all_sorted_indices]
+        predicted_bb = predicted_bb[all_sorted_indices]
+        # pos_threshold是一个用于非背景预测的阈值, 如果有预测目标概率小于pos那个值则被视为背景
+        below_min_idx = (conf < pos_threshold)
+        class_id[below_min_idx] = 0
+        conf[below_min_idx] = 1 - conf[below_min_idx] # 增加背景的置信度
+        # 返回预测的信息，共num_anchors行，每一行6个信息，分别为：类别，预测概率值，4个左上右下坐标
+        pred_info = torch.cat((class_id.unsqueeze(1), conf.unsqueeze(1), predicted_bb), dim=1)
+        out.append(pred_info)
+    return torch.stack(out)
\ No newline at end of file
